import telegram
import time
import os
import shutil

class TelegramBot():

    def __init__(self, token, user_ids, tmp_dir='./temp/'):
        self.bot = telegram.Bot(token)
        self.user_ids = user_ids
        self.tmp_dir = tmp_dir

    def send_text(self, text):
        return self.bot.send_message(self.user_ids, text)

    def update_text(self, message, text):
        if message.text != text:
            try:
                message.edit_text(text)
            except Exception as e:
                pass

    def send_structured_text(self, fields=[], values=[], units=[]):
        msg_text = ''
        for i in range(len(fields)):
            v = values[i] if not isinstance(
                values[i], float) else round(values[i], 3)
            msg_text += fields[i]+': '+str(v)+units[i]+'\n'

        return self.send_text(msg_text)

    def update_structured_text(self, message, fields=[], values=[], units=[]):
        msg_text = ''
        for i in range(len(fields)):
            v = values[i] if not isinstance(
                values[i], float) else round(values[i], 2)
            msg_text += fields[i]+': '+str(v)+units[i]+'\n'

        return self.update_text(message, msg_text)

    def send_plot(self, plt, name=None):
        if not os.path.exists(self.tmp_dir):
            os.makedirs(self.tmp_dir)

        if name is None:
            ts = int(time.time())
            img_path = self.tmp_dir+str(ts)+'.png'
        else:
            img_path = self.tmp_dir+name

        plt.savefig(img_path, dpi=100)
        return self.bot.send_photo(self.user_ids, open(img_path, 'rb'))

    def send_image(self, img_path):
        return self.bot.send_photo(self.user_ids, open(img_path, 'rb'))

    def update_plot(self, message, plt, name=None):
        if not os.path.exists(self.tmp_dir):
            os.makedirs(self.tmp_dir)

        if name is None:
            ts = int(time.time())
            img_path = self.tmp_dir+str(ts)+'.png'
        else:
            img_path = self.tmp_dir+name

        plt.savefig(img_path, dpi=100)
        new_media = telegram.InputMediaPhoto(open(img_path, 'rb'))
        try:
            message.edit_media(new_media)
        except Exception as e:
            pass

    def clean_tmp_dir(self):
        shutil.rmtree(self.tmp_dir)


import time
import keras
from datetime import timedelta
import matplotlib.pyplot as plt
import numpy as np
from .bot import TelegramBot
from .utils import TelegramTqdm


class KerasTelegramCallback(keras.callbacks.Callback):
    def __init__(self, bot: TelegramBot, epoch_bar=True, to_plot=[]):
        self.bot = bot
        self.name = str(int(time.time()))+'.png'
        self.epoch_bar = epoch_bar
        if self.epoch_bar:
            self.pbar = None
        self.to_plot = to_plot
        self.plot_id = {}

        for i in range(len(self.to_plot)):
            p = self.to_plot[i]
            p['id'] = i
            self.plot_id[p['id']] = None

    def on_train_begin(self, logs={}):
        self.n_epochs = self.params['epochs']
        self.metrics = self.params['metrics']

        self.n_steps = self.params['samples']//self.params['batch_size']
        self.n_steps += 1 if self.params['samples'] % self.params['batch_size'] != 0 else 0

        self.history = {}
        for metric in self.metrics:
            self.history[metric] = []

        self.current_epoch = 0

        fields = ['Status', 'Epoch']
        units = ['', '']
        values = ['TRAINING', str(self.current_epoch)+'/'+str(self.n_epochs)]

        self.msg = self.bot.send_structured_text(fields, values, units)

    def on_batch_begin(self, batch, logs={}):
        if self.epoch_bar:
            if self.pbar is None:
                self.obj = TelegramTqdm(self.bot)
                self.pbar = self.obj(total=self.n_steps)

    def on_batch_end(self, batch, logs={}):
        if self.epoch_bar:
            message = ''
            for m in self.metrics:
                if 'val_' not in m:
                    message += m+(': %.4f - ' % logs[m])

            self.pbar.set_description(message[:-3])
            self.pbar.update(1)

    def on_epoch_end(self, batch, logs={}):
        self.current_epoch += 1

        fields = ['Status', 'Epoch']
        units = ['', '']
        values = ['TRAINING', str(self.current_epoch)+'/'+str(self.n_epochs)]

        self.bot.update_structured_text(self.msg, fields, values, units)

        for m in self.metrics:
            self.history[m].append(logs[m])

        for plot_par in self.to_plot:
            self.plot_id[plot_par['id']] = self.plot(
                plot_par, self.plot_id[plot_par['id']])

    def on_train_end(self, logs={}):
        fields = ['Status']
        units = ['']
        values = ['TRAINING END']

        self.bot.update_structured_text(self.msg, fields, values, units)

    def plot(self, params: dict,
             plot_id=None):

        metrics = params['metrics']
        title = params.get('title', 'Loss')
        ylabel = params.get('ylabel', 'loss')
        xlabel = params.get('xlabel', '# Epochs')
        xlim = params.get('xlim', None)
        ylim = params.get('ylim', None)

        t = [k+1 for k in range(len(self.history[metrics[0]]))]

        for m in metrics:
            plt.plot(t, self.history[m])

        plt.title(title)
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.legend(metrics)

        if xlim is not None:
            plt.xlim(xlim[0], xlim[1])
        if ylim is not None:
            plt.ylim(ylim[0], ylim[1])

        if plot_id is None:
            plot_id = self.bot.send_plot(plt, self.name)
        else:
            self.bot.update_plot(plot_id, plt, self.name)

        plt.gcf().clear()
        return plot_id



from datetime import datetime
from tqdm import tqdm


class _TelegramIO():
    def __init__(self, bot, show_last_update=True):
        self.bot = bot
        self.text = self.prev_text = '<< Init tqdm >>'
        self.message = self.bot.send_text(self.text)
        self.show_last_update = show_last_update

    def write(self, s):
        new_text = s.strip().replace('\r', '')
        if len(new_text) != 0:
            self.text = new_text

    def flush(self):
        if self.prev_text != self.text:
            if '%' in self.text:
                self.bot.update_text(self.message, self.text +
                                     '\nLast update: {}'.format(datetime.now()) if self.show_last_update else self.text)
                self.prev_text = self.text


class TelegramTqdm():
    def __init__(self, bot, show_last_update=False):
        self.bot = bot
        self.tg_io = _TelegramIO(self.bot, show_last_update)

    def __call__(self, iterable=None, show_last_update=False,
                 desc=None, total=None, leave=True, ncols=None, mininterval=1.0, maxinterval=10.0,
                 miniters=None, ascii=False, disable=False, unit='it',
                 unit_scale=False, dynamic_ncols=False, smoothing=0.3,
                 bar_format=None, initial=0, position=None, postfix=None,
                 unit_divisor=1000, gui=False, **kwargs):

        params = {
            'desc': desc,
            'total': total,
            'leave': leave,
            'file': self.tg_io,
            'ncols': ncols,
            'mininterval': mininterval,
            'maxinterval': maxinterval,
            'miniters': miniters,
            'ascii': ascii,
            'disable': disable,
            'unit': unit,
            'unit_scale': unit_scale,
            'dynamic_ncols': dynamic_ncols,
            'smoothing': smoothing,
            'bar_format': bar_format,
            'initial': initial,
            'position': position,
            'postfix': postfix,
            'unit_divisor': unit_divisor,
            'gui': gui
        }

        params.update(kwargs)

        if iterable is not None:
            params['iterable'] = iterable
            
        return tqdm(**params)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bob-telegram-tools",  # Replace with your own username
    version="1.0.0",
    author="Antonio Roberto",
    author_email="roberto.antonio@outlook.it",
    description="A package to monitor your Machine Learning trainings every where without any additional app.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/antonioroberto1994/bob_telegram_tools",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.9.2',
        'matplotlib>=2.0.0',
        'tqdm>=4.11.2',
        'python-telegram-bot'
    ],
)
